from functools import partial
from multiprocessing import Process, get_context
from pathlib import Path

from hugedict.hugedict.rocksdb import Options, SecondaryDB, primary_db, stop_primary_db


def test_start_and_close_primary_db(wdprops: Path, url: str):
    p = get_context("spawn").Process(
        target=primary_db, args=(url, str(wdprops), Options())
    )
    p.start()

    # # wait till primary db is ready
    # time.sleep(1)

    assert p.exitcode is None

    # stop it
    stop_primary_db(url)

    # wait 3 seconds
    p.join(3)

    assert p.exitcode == 0


def test_primary_secondary_db(wdprops: Path, url: str):
    p = get_context("spawn").Process(
        target=primary_db, args=(url, str(wdprops), Options())
    )
    p.start()

    db0 = SecondaryDB(
        url,
        str(wdprops),
        str(wdprops / "secondaries/0"),
        Options(),
        deser_value=partial(str, encoding="utf-8"),
        ser_value=str.encode,
    )

    db1 = SecondaryDB(
        url,
        str(wdprops),
        str(wdprops / "secondaries/1"),
        Options(),
        deser_value=partial(str, encoding="utf-8"),
        ser_value=str.encode,
    )

    assert "P58" not in db0
    assert "P58" not in db1

    # setting P58 on a secondary db will make it available to other db through the primary db
    db1["P58"] = "screenwriter"
    assert "P58" in db0
    assert "P58" in db1

    stop_primary_db(url)
    p.join(3)
    assert p.exitcode == 0



import shutil
from hugedict.v1.rocksdb import RocksDBDict
from hugedict.v1.mrsw_rocksdb import *
from multiprocessing.pool import *
from hugedict.v1.parallel import MyManager

shutil.rmtree("/tmp/hugedict/test.db", ignore_errors=True)

db = RocksDBDict("/tmp/hugedict/test.db", create_if_missing=True)
db["a"] = "b"

print(">>> 1", db["a"])

# input("enter to close")

# input("enter to continue")
# with MyManager() as manager:
#     db.close()

#     db2 = manager.PrimarySyncedRocksDBDict(  # type: ignore
#         db.dbpath,
#         create_if_missing=True,
#     )
manager = MyManager()
try:
    db.close()
    manager.start()

    db2 = manager.PrimarySyncedRocksDBDict(  # type: ignore
        db.dbpath,
        create_if_missing=True,
    )
finally:
    manager.shutdown()
# db2 = RocksDBDict("/tmp/hugedict/test.db", create_if_missing=True)

# db2["a"] = "b"
# print(">>> 2", db2.get("a"))



from hugedict.hugedict.rocksdb import (
    RocksDBDict,
    Options as RocksDBOptions,
    load as rocksdb_load,
    fixed_prefix,
    fixed_prefix_alike,
)
from hugedict.parallel import (
    Parallel,
    Compressing as ParallelCacheCompressingMode,
    CacheFnKey,
)
from hugedict.types import HugeMapping, HugeMutableMapping
from hugedict.chained_mapping import ChainedMapping
from hugedict.cachedict import CacheDict

__all__ = [
    "RocksDBDict",
    "RocksDBOptions",
    "rocksdb_load",
    "fixed_prefix",
    "fixed_prefix_alike",
    "Parallel",
    "ParallelCacheCompressingMode",
    "CacheFnKey",
    "HugeMapping",
    "HugeMutableMapping",
    "ChainedMapping",
    "CacheDict",
]



from __future__ import annotations
from contextlib import contextmanager
import enum
from functools import partial
import gc
import gzip
from inspect import signature
from multiprocessing import Pool, Process, get_context
from multiprocessing.context import SpawnProcess
from multiprocessing.pool import ThreadPool
from operator import itemgetter
import os
from pathlib import Path
import pickle
import random
import shutil
import time
from typing import Any, Callable, List, Literal, Optional, Union, MutableMapping, cast
from uuid import uuid4

from loguru import logger
import orjson
from tqdm import tqdm
from hugedict.types import F, Fn
from hugedict.misc import (
    Chain2,
    compress_zstd6_pyobject,
    decompress_zstd_pyobject,
    identity,
    compress_pyobject,
    decompress_pyobject,
)
from hugedict.hugedict.rocksdb import (
    Options,
    RocksDBDict,
    SecondaryDB,
    primary_db,
    stop_primary_db,
)


class Compressing(enum.Flag):
    NoCompression = enum.auto()
    CompressKey = enum.auto()
    CompressValue = enum.auto()
    CompressAll = CompressValue | CompressKey


class Parallel:
    SWITCHED_MRSW_CACHE = False

    """Utility to execute a function over a list of inputs in parallel. It is useful
    when you have an expensive function and want to cache it results.

    Arguments:
        start_method: method used to start the process. Using fork is recommended if
            the code do not involve any lock (you may encounter deadlock).
            Note: spawn mode is currently not working.
    """

    def __init__(self, start_method: Literal["fork", "spawn", "forkserver"] = "fork"):
        self.start_method = start_method
        self.caches: List[LazyDBCacheFn] = []

    @contextmanager
    def init_cache(self):
        try:
            for cache in self.caches:
                if cache.db is not None:
                    del cache.db
                    cache.db = None
            gc.collect()
            yield
        finally:
            for cache in self.caches:
                if cache.db is not None:
                    del cache.db
                    cache.db = None
            gc.collect()

    @contextmanager
    def switch_mrsw_cache(self):
        primaries = []
        for cache in self.caches:
            primaries.append(cache.start_primary())

        yield

        for cache, primary in zip(self.caches, primaries):
            if primary is not None:
                cache.stop_primary()

        for primary in primaries:
            primary.join(30)
            assert primary.exitcode == 0

        for cache in self.caches:
            cache.cleanup()

    @staticmethod
    def spawn_init_cache(enable_mrsw: bool):
        Parallel.SWITCHED_MRSW_CACHE = enable_mrsw

    def cache_func(
        self,
        dbpath: Union[Path, str],
        namespace: str = "",
        compress=Compressing.NoCompression,
        key: Optional[Callable[[str, tuple, dict], bytes]] = None,
        url: Optional[str] = None,
    ) -> Callable[[F], F]:
        """Cache a function (only work when using with Parallel object)"""
        for cache in self.caches:
            assert cache.dbpath != Path(dbpath), "dbpath must be unique"

        def wrapper_fn(func: F) -> F:
            if compress & Compressing.CompressKey:
                if key is None:
                    keyfn = compress_pyobject
                else:
                    keyfn = Chain2(partial(gzip.compress, mtime=0), key)
            else:
                keyfn = key

            if compress & Compressing.CompressValue:
                ser_value, deser_value = (
                    compress_zstd6_pyobject,
                    decompress_zstd_pyobject,
                )
            else:
                ser_value, deser_value = pickle.dumps, pickle.loads

            cache = LazyDBCacheFn(
                dbpath=Path(dbpath),
                dbopts=Options(create_if_missing=True),
                deser_value=deser_value,
                ser_value=ser_value,
                fn=func,
                namespace=namespace,
                key=keyfn,  # type: ignore
                url=url,
            )

            self.caches.append(cache)
            return cache.run  # type: ignore

        return wrapper_fn

    def map(
        self,
        fn: Fn,
        inputs: list,
        show_progress=False,
        progress_desc="",
        is_parallel=True,
        use_threadpool=False,
        n_processes: Optional[int] = None,
        ignore_error: bool = False,
    ) -> List[Any]:
        """Execute a map function over each input in parallel"""
        with self.init_cache():
            if not is_parallel:
                wrapped_fn = ParallelFnWrapper(fn, ignore_error=False).run
                return [
                    wrapped_fn((i, item))[1]
                    for i, item in tqdm(
                        enumerate(inputs),
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    )
                ]

            if use_threadpool:
                # it won't break rocksdb when using threadpool because of GIL.
                with ThreadPool(processes=n_processes) as pool:
                    iter = pool.imap_unordered(
                        ParallelFnWrapper(fn, ignore_error).run,
                        enumerate(inputs),
                    )
                    results = []
                    for result in tqdm(
                        iter,
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    ):
                        results.append(result)
                    results.sort(key=itemgetter(0))
            else:
                # have to switch to multi read single write mode
                with self.switch_mrsw_cache():
                    if self.start_method == "fork":
                        initializer = None
                        initargs = ()
                    else:
                        initializer = Parallel.spawn_init_cache
                        initargs = (True,)

                    # start a pool of processes
                    with get_context(self.start_method).Pool(
                        processes=n_processes,
                        initializer=initializer,
                        initargs=initargs,
                    ) as pool:
                        iter = pool.imap_unordered(
                            ParallelFnWrapper(fn, ignore_error).run,
                            enumerate(inputs),
                        )
                        results = []
                        for result in tqdm(
                            iter,
                            total=len(inputs),
                            desc=progress_desc,
                            disable=not show_progress,
                        ):
                            results.append(result)
                        results.sort(key=itemgetter(0))

        return [v for i, v in results]

    def foreach(
        self,
        fn: Fn,
        inputs: list,
        show_progress=False,
        progress_desc="",
        is_parallel=True,
        use_threadpool=False,
        n_processes: Optional[int] = None,
        ignore_error: bool = False,
    ) -> None:
        """Execute a map function over each input in parallel"""
        with self.init_cache():
            if not is_parallel:
                for item in tqdm(
                    inputs,
                    total=len(inputs),
                    desc=progress_desc,
                    disable=not show_progress,
                ):
                    fn(item)
                return

            if use_threadpool:
                # it won't break rocksdb when using threadpool because of GIL.
                with ThreadPool(processes=n_processes) as pool:
                    iter = pool.imap_unordered(
                        ParallelFnWrapper(fn, ignore_error).run_no_return,
                        enumerate(inputs),
                    )
                    results = []
                    for result in tqdm(
                        iter,
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    ):
                        results.append(result)
                    return

            # have to switch to multi read single write mode
            with self.switch_mrsw_cache():
                # start a pool of processes
                if self.start_method == "fork":
                    initializer = None
                    initargs = ()
                else:
                    initializer = Parallel.spawn_init_cache
                    initargs = (True,)

                # start a pool of processes
                with get_context(self.start_method).Pool(
                    processes=n_processes,
                    initializer=initializer,
                    initargs=initargs,
                ) as pool:
                    iter = pool.imap_unordered(
                        ParallelFnWrapper(fn, ignore_error).run_no_return,
                        enumerate(inputs),
                    )
                    results = []
                    for result in tqdm(
                        iter,
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    ):
                        results.append(result)
                    return


class ParallelFnWrapper:
    def __init__(self, fn: Fn, ignore_error=False):
        self.fn = fn
        # method of instances won't have `self` parameter, so we do not need to worry about it
        if hasattr(fn, "__self__") and isinstance(fn.__self__, LazyDBCacheFn):
            fn_params = signature(fn.__self__.fn).parameters
        else:
            fn_params = signature(fn).parameters
        self.spread_fn_args = len(fn_params) > 1
        self.ignore_error = ignore_error

    def run(self, args):
        idx, r = args
        try:
            if self.spread_fn_args:
                r = self.fn(*r)
            else:
                r = self.fn(r)
            return idx, r
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            if self.ignore_error:
                return idx, None
            raise

    def run_no_return(self, args):
        idx, r = args
        try:
            if self.spread_fn_args:
                r = self.fn(*r)
            else:
                r = self.fn(r)
            return idx, None
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            if self.ignore_error:
                return idx, None
            raise


class LazyDBCacheFn:
    SECONDARY_DIR = "secondaries"

    def __init__(
        self,
        dbpath: Path,
        dbopts: Options,
        deser_value: Callable[[memoryview], Any],
        ser_value: Callable[[Any], bytes],
        fn: Fn,
        key: Optional[Callable[[str, tuple, dict], bytes]] = None,
        namespace: str = "",
        url: Optional[str] = None,
    ):
        self.dbopts = dbopts
        self.dbpath = dbpath
        self.deser_value = deser_value
        self.ser_value = ser_value
        self.db: Optional[MutableMapping[bytes, Any]] = None
        self.is_mrsw = False
        self.namespace = namespace

        self.fn = fn
        self.fn_name = (
            namespace + ":" + fn.__name__ if len(namespace) > 0 else fn.__name__
        )
        self.key = key or CacheFnKey.default_key

        self.url = url or f"ipc://{self.dbpath.absolute()}/hugedict.ipc"

    def run(self, *args, **kwargs):
        key = self.key(self.fn_name, args, kwargs)
        if self.db is None:
            if not self.is_mrsw and not Parallel.SWITCHED_MRSW_CACHE:
                self.db = RocksDBDict(
                    str(self.dbpath),
                    self.dbopts,
                    bytes,
                    self.deser_value,
                    self.ser_value,
                )
            else:
                (self.dbpath / self.SECONDARY_DIR).mkdir(parents=True, exist_ok=True)
                self.db = cast(
                    MutableMapping,
                    SecondaryDB(
                        self.url,
                        str(self.dbpath),
                        str(self.dbpath / self.SECONDARY_DIR / str(uuid4())),
                        self.dbopts,
                        self.deser_value,
                        self.ser_value,
                    ),
                )

        if key not in self.db:
            output = self.fn(*args, **kwargs)
            self.db[key] = output
            return output

        return self.db[key]

    def start_primary(self) -> Optional[SpawnProcess]:
        assert (
            self.db is None
        ), "Switching to Multi-read single-write mode must done before starting a primary instance"
        self.is_mrsw = True
        self.cleanup()

        p = get_context("spawn").Process(
            target=primary_db, args=(self.url, str(self.dbpath), self.dbopts)
        )
        p.start()
        return p

    def stop_primary(self):
        self.is_mrsw = False
        stop_primary_db(self.url)

    def cleanup(self):
        # clean previous directory if exists
        if (self.dbpath / self.SECONDARY_DIR).exists():
            shutil.rmtree(self.dbpath / self.SECONDARY_DIR)


class CacheFnKey:
    @staticmethod
    def default_key(fn_name, args, kwargs):
        return orjson.dumps((fn_name, args, kwargs))

    @staticmethod
    def single_str_arg(fn_name: str, args: tuple, kwargs: dict):
        return args[0].encode()



import os
from typing import Generator
from uuid import uuid4
import pytest
from pathlib import Path


@pytest.fixture(scope="session")
def test_dir() -> Path:
    """Get the testing directory"""
    return Path(__file__).absolute().parent.parent


@pytest.fixture(scope="session")
def resource_dir() -> Path:
    """Get the testing directory"""
    return Path(__file__).absolute().parent.parent / "resources"


@pytest.fixture
def url() -> Generator[str, None, None]:
    file = f"/tmp/{str(uuid4())}.ipc"
    url = f"ipc://{file}"

    assert not Path(file).exists()
    yield url

    if Path(file).exists():
        os.remove(file)



from multiprocessing.dummy import freeze_support
import os
import multiprocessing as mp


x = 0


def f():
    global x
    print("In process", os.getpid(), "x =", x)


def main():
    global x
    x = 1

    f()

    p = mp.Process(target=f)
    p.start()
    p.join()


if __name__ == "__main__":
    mp.set_start_method("spawn")
    # freeze_support()
    main()



"""
Download the mapping from wikidata property ids to their labels.

Prerequisite:
    pip install web-table-extractor
"""


from pathlib import Path
from table_extractor.fetch_table import fetch_tables
from tqdm import tqdm

pages = [
    "https://www.wikidata.org/wiki/Wikidata:List_of_properties/human",
    "https://www.wikidata.org/wiki/Wikidata:List_of_properties/graph",
    "https://www.wikidata.org/wiki/Wikidata:List_of_properties/e-commerce",
]

outdir = Path(__file__).parent

for page in tqdm(pages):
    outfile = outdir / (page.rsplit("/", 1)[-1] + ".tsv")
    tables = fetch_tables(page)
    map = {}

    for table in tables:
        header = table.rows[0]
        if (
            all(c.is_header for c in header.cells)
            and header.cells[0].value.strip() == "Title"
            and header.cells[1].value.strip() == "ID"
        ):
            for row in table.rows[1:]:
                title = row.cells[0].value.strip()
                id = row.cells[1].value.strip()
                map[id] = title

    with open(outfile, "w") as f:
        for id, title in map.items():
            f.write(f"{id}\t{title}\n")



from multiprocessing import freeze_support, set_start_method, get_start_method
import os
import time
import shutil

from hugedict.parallel import Parallel

shutil.rmtree("/tmp/hugedict/test.db", ignore_errors=True)


pp = Parallel()


@pp.cache_func("/tmp/hugedict/test.db")
def heavy_computing(seconds: float):
    time.sleep(seconds)
    return seconds * 2


def run(seconds: float):
    return heavy_computing(seconds)


if __name__ == "__main__":
    # print("@@", get_start_method())
    # set_start_method("fork")
    # freeze_support()

    output = pp.map(
        run,
        [0.5, 1, 0.7, 0.3, 0.6],
        n_processes=3,
        # use_threadpool=True,
        # is_parallel=False,
    )

    assert output == [1, 2, 1.4, 0.6, 1.2]



# Prerequisites:
# - install sem-desc: `pip install sem-desc`
# - data to test with in the ./data directory (e.g., ./data/wdentities/part-*.tsv.gz)
from functools import partial
import random
from loguru import logger
from matplotlib.style import available
import orjson
from dataclasses import dataclass
import multiprocessing
import shutil
from pathlib import Path
from typing import Optional

from hugedict.v1.loader import FileFormat, load
from rocksdb import DB, Options  # type: ignore
from sm.prelude import M

from hugedict.v1.rocksdb import RocksDBDict
from hugedict.hugedict.rocksdb import RocksDBDict as RustRocksDBDict, Options

bench_dir = Path(__file__).parent

if not (bench_dir / "tempdir" / "keys.txt.gz").exists():
    db = RocksDBDict(
        str(bench_dir / "tempdir" / "python-rocksdb.db"), deser_value=orjson.loads
    )

    keys = list(db.keys())
    M.serialize_lines(keys, bench_dir / "tempdir" / "keys.txt.gz")
    logger.info("Stop after creating list of keys. You need to rerun the script")
    exit(0)

keys = M.deserialize_lines(bench_dir / "tempdir" / "keys.txt.gz", trim=True)
random.shuffle(keys)
keys = keys[:3000]

db = RocksDBDict(
    str(bench_dir / "tempdir" / "python-rocksdb.db"), deser_value=orjson.loads
)
with M.Timer().watch_and_report("python-rocksdb"):
    count = 0
    for id in keys:
        count += len(db[id])

db2 = RustRocksDBDict(
    path=str(bench_dir / "tempdir" / "rust-rocksdb.db"),
    options=Options(),
    deser_key=partial(str, encoding="utf-8"),
    deser_value=orjson.loads,
    ser_value=orjson.dumps,
)
with M.Timer().watch_and_report("rust-rocksdb"):
    count = 0
    for id in keys:
        count += len(db2[id])

print("Done")



from functools import partial
from pathlib import Path
from typing import List, Mapping
import pytest
from hugedict.cachedict import CacheDict
from hugedict.hugedict.rocksdb import RocksDBDict, Options, fixed_prefix_alike
from tests.python.test_mapping import TestMutableMappingSuite


class TestRocksDBDict(TestMutableMappingSuite):
    DELETE_MISSING_KEY_RAISE_ERROR = False

    @pytest.fixture
    def mapping(self, wdprops: Path) -> Mapping:
        return RocksDBDict(
            str(wdprops),
            Options(),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=partial(str, encoding="utf-8"),
            ser_value=str.encode,
        )

    @pytest.fixture
    def size(self) -> int:
        return 171

    @pytest.fixture
    def existed_items(self) -> List[tuple]:
        return [
            ("P7462", "graph diameter"),
            ("P7166", "Acharts.co song ID"),
            ("P405", "taxon author"),
            ("P2842", "place of marriage"),
            ("P2937", "parliamentary term"),
            ("P7957", "Directory of Maîtres d'art"),
        ]

    @pytest.fixture
    def new_items(self) -> List[tuple]:
        return [
            ("P58", "screenwriter"),
            ("P106", "occupation"),
            ("P123", "publisher"),
            ("P162", "producer"),
            ("P170", "creator"),
            ("P175", "performer"),
        ]

    @pytest.fixture
    def unknown_keys(self) -> List:
        return ["Q10", 5, b"whatever", None]

    def test_cache(
        self,
        mapping: RocksDBDict,
        existed_items: List[tuple],
        new_items: List[tuple],
        unknown_keys: List,
    ):
        mcache = mapping.cache()
        assert isinstance(mcache, CacheDict)

        for k, v in existed_items:
            assert k in mapping
            assert k in mcache

        for k in [x[0] for x in new_items] + unknown_keys:
            assert k not in mapping
            assert k not in mcache

    def test_has_properties(self, mapping: RocksDBDict):
        assert hasattr(mapping, "deser_value")
        assert hasattr(mapping, "ser_value")

    def test_estimate_num_keys(self, mapping: RocksDBDict):
        # we create database from scratch, we haven't update/delete any key so the
        # estimation should be correct
        assert mapping.get_int_property("rocksdb.estimate-num-keys") == len(mapping)

    def test__put(self, mapping: RocksDBDict, new_items: List[tuple]):
        for k, v in new_items:
            mapping._put(k.encode(), v.encode())

        for k, v in new_items:
            assert mapping[k] == v

    def test_open_readonly(
        self, wdprops: Path, existed_items: List[tuple], new_items: List[tuple]
    ):
        db = RocksDBDict(
            str(wdprops),
            Options(),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=partial(str, encoding="utf-8"),
            ser_value=str.encode,
            readonly=True,
        )

        for k, v in existed_items:
            assert db[k] == v

        # writing to db in readonly mode is not allowed
        for k, v in new_items:
            with pytest.raises(RuntimeError):
                db[k] = v

    def test_open_secondary_mode(
        self,
        wdprops: Path,
        tmp_path: Path,
        existed_items: List[tuple],
        new_items: List[tuple],
    ):
        primary_db = RocksDBDict(
            str(wdprops),
            Options(),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=partial(str, encoding="utf-8"),
            ser_value=str.encode,
        )

        secondary_db = RocksDBDict(
            str(wdprops),
            Options(),
            deser_key=partial(str, encoding="utf-8"),
            deser_value=partial(str, encoding="utf-8"),
            ser_value=str.encode,
            secondary_mode=True,
            secondary_path=str(tmp_path),
        )

        # writing to secondary db is not allowed
        for k, v in new_items:
            with pytest.raises(RuntimeError):
                secondary_db[k] = v

        # but writing to the primary db is okay
        for k, v in new_items:
            primary_db[k] = v

        # the new key is not available in the secondary db until we try to catch up
        for k, v in new_items:
            assert k not in secondary_db
        secondary_db.try_catch_up_with_primary()
        for k, v in new_items:
            assert secondary_db[k] == v

    def test_seek_iterators(self, wdprops: Path):
        deser_key = partial(str, encoding="utf-8")
        deser_value = partial(str, encoding="utf-8")
        ser_value = str.encode

        get_db = lambda opts: RocksDBDict(
            str(wdprops),
            opts,
            deser_key=deser_key,
            deser_value=deser_value,
            ser_value=ser_value,
        )

        # default opts without prefix_extractor will also yield keys not matched the entired prefix
        db = get_db(Options())

        prefix_counter = 0
        nonprefix_counter = 0

        subitems = dict(db.seek_items("P7"))
        for k in db.seek_keys("P7"):
            if k.startswith("P7"):
                prefix_counter += 1
            else:
                nonprefix_counter += 1
            assert k in subitems and subitems.pop(k) == db[k]

        assert len(subitems) == 0
        assert prefix_counter > 0
        assert nonprefix_counter > 0
        assert prefix_counter + nonprefix_counter < len(db)

        del db

        # get correct results with prefix_extractor
        db = get_db(
            Options(
                prefix_extractor=fixed_prefix_alike(
                    type="fixed_prefix_alike", prefix="P7"
                )
            )
        )
        prefix_counter = 0
        subitems = dict(db.seek_items("P7"))
        for k in db.seek_keys("P7"):
            assert k.startswith("P7")
            prefix_counter += 1
            assert k in subitems and subitems.pop(k) == db[k]
        assert prefix_counter > 0
        assert len(subitems) == 0



from abc import ABC, abstractmethod
import os
from pathlib import Path
import pickle
from typing import (
    Callable,
    Dict,
    Iterator,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    MutableMapping,
)
from hugedict.v1.rocksdb import RocksDBDict
from hugedict.types import K, V, InvalidUsageError
from pybloomfilter import BloomFilter


BloomFilterArgs = NamedTuple(
    "BloomFilterArgs", [("capacity", int), ("error_rate", float)]
)
DEFAULT_BLOOMFILTER_ARGS = (int(1e6), 0.001)


class IProxyDict(ABC):
    @abstractmethod
    def raw_has(self, key: bytes) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def raw_get(self, key: bytes) -> Optional[bytes]:
        raise NotImplementedError()

    @abstractmethod
    def raw_set(self, key: bytes, value: bytes):
        raise NotImplementedError()


class PrimarySyncedRocksDBDict(RocksDBDict[K, V], IProxyDict):
    def __init__(
        self,
        dbpath: Union[Path, str],
        create_if_missing=False,
        deser_key: Callable[[bytes], K] = None,
        deser_value: Callable[[bytes], V] = None,
        ser_key: Callable[[K], bytes] = None,
        ser_value: Callable[[V], bytes] = None,
        bloomfilter_args: BloomFilterArgs = None,
        enable_bloomfilter: bool = True,
    ) -> None:
        super().__init__(
            dbpath=dbpath,
            create_if_missing=create_if_missing,
            deser_key=deser_key,
            deser_value=deser_value,
            ser_key=ser_key,
            ser_value=ser_value,
        )

        # create a default of 1M keys
        self.bloomfilter_args = bloomfilter_args or DEFAULT_BLOOMFILTER_ARGS
        self.enable_bloomfilter = enable_bloomfilter
        if self.enable_bloomfilter:
            self.bloomfilter = BloomFilter(
                *self.bloomfilter_args, os.path.join(self.dbpath, "bloomfilter")
            )

    def __setitem__(self, key: K, value: V):
        serkey = self.ser_key(key)
        self.db.put(serkey, self.ser_value(value))
        if self.enable_bloomfilter:
            self.bloomfilter.add(serkey)

    def __delitem__(self, key: K):
        serkey = self.ser_key(key)
        self.db.delete(serkey)
        if self.enable_bloomfilter:
            self.bloomfilter.remove(serkey)

    def pop(self, key: K, default=None):
        serkey = self.ser_key(key)
        item = self.db.get(serkey)
        if item is None:
            return default
        self.db.delete(serkey)
        if self.enable_bloomfilter:
            self.bloomfilter.remove(serkey)
        return self.deser_value(item)

    def raw_has(self, key: bytes) -> bool:
        return self.db.get(key) is not None

    def raw_get(self, key: bytes) -> Optional[bytes]:
        return self.db.get(key)

    def raw_set(self, key: bytes, value: bytes):
        self.db.put(key, value)
        if self.enable_bloomfilter:
            self.bloomfilter.add(key)


class SecondarySyncedRocksDBDict(RocksDBDict[K, V]):
    def __init__(
        self,
        primary: IProxyDict,
        dbpath: Union[Path, str],
        secondary_name: str,
        deser_key: Optional[Callable[[bytes], K]] = None,
        deser_value: Optional[Callable[[bytes], V]] = None,
        ser_key: Optional[Callable[[K], bytes]] = None,
        ser_value: Optional[Callable[[V], bytes]] = None,
        enable_bloomfilter: bool = True,
    ):
        super().__init__(
            dbpath=dbpath,
            secondary_name=secondary_name,
            create_if_missing=False,
            read_only=False,
            deser_key=deser_key,
            deser_value=deser_value,
            ser_key=ser_key,
            ser_value=ser_value,
        )

        self.primary = primary
        # create a default of 1M keys
        self.enable_bloomfilter = enable_bloomfilter
        if self.enable_bloomfilter:
            self.bloomfilter = BloomFilter.open(
                os.path.join(self.dbpath, "bloomfilter"), "r"
            )

    def __getitem__(self, key: K) -> V:
        serkey = self.ser_key(key)
        item = self.db.get(serkey)

        if item is not None:
            return self.deser_value(item)

        if self.enable_bloomfilter and serkey not in self.bloomfilter:  # type: ignore
            # can't find the key anywhere
            raise KeyError(key)

        item = self.primary.raw_get(serkey)
        if item is None:
            raise KeyError(key)
        return self.deser_value(item)

    def __setitem__(self, key: K, value: V):
        self.primary.raw_set(self.ser_key(key), self.ser_value(value))

    def __delitem__(self, key: K):
        raise InvalidUsageError()

    def __iter__(self) -> Iterator[K]:
        raise InvalidUsageError()

    def __len__(self):
        raise InvalidUsageError()

    def __contains__(self, key):
        serkey = self.ser_key(key)
        return (
            (self.db.get(serkey) is not None)
            or (
                self.enable_bloomfilter and (serkey in self.bloomfilter) and self.primary.raw_has(serkey)  # type: ignore
            )
            or (not self.enable_bloomfilter and self.primary.raw_has(serkey))
        )

    def keys(self) -> Iterator[K]:
        raise InvalidUsageError()

    def values(self) -> Iterator[V]:
        raise InvalidUsageError()

    def get(self, key: K, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def pop(self, key: K, default=None):
        raise InvalidUsageError()



from __future__ import annotations
from typing import Callable, Mapping, MutableMapping, TypeVar, Union
from abc import abstractmethod


# generic parameters
K = TypeVar("K")
KP = TypeVar("KP", bound=Union[str, bytes, int, None])
V = TypeVar("V")
T = TypeVar("T")
F = TypeVar("F", bound=Callable)
Fn = Callable


class InvalidUsageError(Exception):
    pass


class HugeMapping(Mapping[K, V]):
    @abstractmethod
    def cache(self) -> HugeMapping[K, V]:
        """Return a new mapping that will cache the results to
        avoid calling to an external mapping
        """
        pass


class HugeMutableMapping(MutableMapping[K, V]):
    @abstractmethod
    def cache(self) -> HugeMutableMapping[K, V]:
        """Return a new mapping that will cache the results to
        avoid calling to an external mapping
        """
        pass



import os
from tempfile import TemporaryDirectory
import orjson

from multiprocessing import Process, Pipe
import pybloomfilter


def process1(conn):
    msg = orjson.loads(conn.recv())
    assert msg["status"] == "start"
    bloomfilter = pybloomfilter.BloomFilter(10000, 0.1, msg["bloomfile"])
    bloomfilter.add(msg["key1"])
    conn.send(orjson.dumps({"status": "done", "has_key": msg["key1"] in bloomfilter}))

    msg = orjson.loads(conn.recv())
    assert msg["status"] == "start"
    bloomfilter.add(msg["key2"])
    conn.send(orjson.dumps({"status": "done", "has_key": msg["key2"] in bloomfilter}))


def process2(conn):
    msg = orjson.loads(conn.recv())
    assert msg["status"] == "start"
    bloomfilter = pybloomfilter.BloomFilter.open(msg["bloomfile"], "r")
    conn.send(orjson.dumps({"status": "done", "has_key": msg["key1"] in bloomfilter}))

    msg = orjson.loads(conn.recv())
    assert msg["status"] == "start"
    conn.send(orjson.dumps({"status": "done", "has_key": msg["key2"] in bloomfilter}))


def test_bloomfilter_shared_data():
    with TemporaryDirectory() as tempdir:
        bloomfile = os.path.join(tempdir, "bloomfilter")

        p1conn_pp, p1conn_cc = Pipe()
        p2conn_pp, p2conn_cc = Pipe()

        p1 = Process(target=process1, args=(p1conn_cc,))
        p1.start()

        p2 = Process(target=process2, args=(p2conn_cc,))
        p2.start()

        key1 = '["heavy_computing",[0.365],{}]'
        key2 = '["heavy_computing",[0.476],{}]'

        p1conn_pp.send(
            orjson.dumps({"status": "start", "bloomfile": bloomfile, "key1": key1})
        )
        assert orjson.loads(p1conn_pp.recv()) == {"status": "done", "has_key": True}

        p2conn_pp.send(
            orjson.dumps({"status": "start", "bloomfile": bloomfile, "key1": key1})
        )
        assert orjson.loads(p2conn_pp.recv()) == {"status": "done", "has_key": True}

        p1conn_pp.send(orjson.dumps({"status": "start", "key2": key2}))
        assert orjson.loads(p1conn_pp.recv()) == {"status": "done", "has_key": True}

        p2conn_pp.send(orjson.dumps({"status": "start", "key2": key2}))
        assert orjson.loads(p2conn_pp.recv()) == {"status": "done", "has_key": True}



from itertools import chain
import os
from typing import Iterator, KeysView, Mapping, MutableMapping, ValuesView
from hugedict.types import K, V


class ChainedMapping(Mapping[K, V]):
    def __init__(self, mapping1: Mapping[K, V], mapping2: Mapping[K, V]) -> None:
        self.mapping1 = mapping1
        self.mapping2 = mapping2

    def __getitem__(self, key: K) -> V:
        if key in self.mapping1:
            return self.mapping1[key]
        return self.mapping2[key]

    def __iter__(self) -> Iterator[K]:
        return chain(iter(self.mapping1), iter(self.mapping2))

    def __len__(self):
        return len(self.mapping1) + len(self.mapping2)

    def __contains__(self, key):
        if key in self.mapping1:
            return True
        return key in self.mapping2

    def get(self, key: K, default=None):
        if key in self.mapping1:
            return self.mapping1[key]
        return self.mapping2.get(key, default)



from typing import Generic, List, Mapping, MutableMapping, Tuple

import pytest


class TestMappingSuite:
    @pytest.fixture
    def mapping(self) -> Mapping:
        return {"P276": "location", "Q5": "human"}

    @pytest.fixture
    def size(self) -> int:
        return 2

    @pytest.fixture
    def existed_items(self) -> List[tuple]:
        return [("Q5", "human")]

    @pytest.fixture
    def unknown_keys(self) -> List:
        return ["Q10", 5, b"whatever"]

    def test_get(
        self,
        mapping: Mapping,
        existed_items: List[tuple],
        unknown_keys: List,
    ):
        for k, v in existed_items:
            assert mapping[k] == v
            if v is not None:
                assert mapping.get(k, None) == v
            else:
                assert mapping.get(k, 1) == v

        for k in unknown_keys:
            with pytest.raises(KeyError):
                mapping[k]

    def test_contains(
        self,
        mapping: Mapping,
        existed_items: List[tuple],
        unknown_keys: List,
    ):
        for k, v in existed_items:
            assert k in mapping

        for k in unknown_keys:
            assert k not in mapping

    def test_len(self, mapping: Mapping, size: int):
        assert len(mapping) == size

    def test_iter(self, mapping: Mapping):
        size = len(mapping)

        keys = list(iter(mapping))
        assert keys == list(mapping.keys()), "Looping order is the same"
        assert len(keys) == size, "Get all keys"
        for k in keys:
            assert k in mapping

        items = dict(mapping.items())
        assert len(items) == size, "Get all items"
        for k, v in items.items():
            assert mapping[k] == v

        values = list(mapping.values())
        assert len(values) == size, "Get all values"
        for i, (k, v) in enumerate(items.items()):
            assert values[i] == v, "Looping order is the same"


class TestMutableMappingSuite(TestMappingSuite):
    DELETE_MISSING_KEY_RAISE_ERROR = True

    @pytest.fixture
    def new_items(self) -> List[tuple]:
        return [("Q8", "happiness")]

    def test_delete(
        self,
        mapping: MutableMapping,
        existed_items: List[tuple],
        new_items: List[tuple],
        unknown_keys: List,
    ):
        size = len(mapping)

        for k in [x[0] for x in new_items] + unknown_keys:
            if self.DELETE_MISSING_KEY_RAISE_ERROR:
                with pytest.raises(KeyError):
                    del mapping[k]
            else:
                del mapping[k]
        assert size == len(mapping)

        for k, v in existed_items:
            del mapping[k]
            assert k not in mapping
        assert size == len(mapping) + len(existed_items)

    def test_pop(
        self,
        mapping: MutableMapping,
        existed_items: List[tuple],
        new_items: List[tuple],
        unknown_keys: List,
    ):
        for k in [x[0] for x in new_items] + unknown_keys:
            assert mapping.pop(k, None) == None

        for k, v in existed_items:
            if v is not None:
                assert mapping.pop(k, None) == v
            else:
                assert mapping.pop(k, 1) == v
            assert k not in mapping

    def test_set(self, mapping: MutableMapping, new_items: List[Tuple]):
        size = len(mapping)
        for k, v in new_items:
            assert k not in mapping
            mapping[k] = v
            assert k in mapping

        assert size + len(new_items) == len(mapping)



import os
from pathlib import Path
import pickle
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    Optional,
    TypeVar,
    Union,
    MutableMapping,
    cast,
)
from hugedict.cachedict import CacheDict
from rocksdb import DB, Options  # type: ignore
from pybloomfilter import BloomFilter
from hugedict.types import K, V


class RocksDBDict(MutableMapping[K, V]):
    """
    A dictionary-like object that uses RocksDB as the underlying storage.

    Args:
        dbpath: location to the rocksdb database folder.
        create_if_missing: If True, create the RocksDB file if it does not exist.
        secondary_name: If provided, open as a secondary instance
        read_only: If True, open the RocksDB file in read-only mode.
        deser_key: A function to deserialize a key from bytes.
        deser_value: A function to deserialize a value from bytes.
        ser_key: A function to serialize a key to bytes.
        ser_value: A function to serialize a value to bytes.
    """

    def __init__(
        self,
        dbpath: Union[Path, str],
        create_if_missing=False,
        secondary_name: Optional[str] = None,
        read_only=False,
        deser_key: Optional[Callable[[bytes], K]] = None,
        deser_value: Optional[Callable[[bytes], V]] = None,
        ser_key: Optional[Callable[[K], bytes]] = None,
        ser_value: Optional[Callable[[V], bytes]] = None,
        db_options: Optional[Dict[str, Any]] = None,
    ):
        if db_options is None:
            db_options = {}
        db_options["create_if_missing"] = create_if_missing
        kwargs: Dict[str, Any] = dict(read_only=read_only)

        if secondary_name is not None:
            if secondary_name == "primary":
                raise ValueError("secondary_name cannot be 'primary'")

            # required to open as secondary instance
            db_options["max_open_files"] = -1
            kwargs["secondary_name"] = os.path.join(dbpath, secondary_name)

        self.dbpath = dbpath
        Path(self.dbpath).mkdir(exist_ok=True, parents=True)
        self.db_options = db_options
        self.kwargs = kwargs
        self.db = DB(
            os.path.join(dbpath, "primary"), Options(**self.db_options), **kwargs
        )
        self.secondary_name = secondary_name
        self.is_primary = secondary_name is None

        self.deser_key = cast(Callable[[bytes], K], deser_key or bytes.decode)
        self.ser_key = cast(Callable[[K], bytes], ser_key or str.encode)

        self.deser_value = cast(Callable[[bytes], V], deser_value or pickle.loads)
        self.ser_value = cast(Callable[[V], bytes], ser_value or pickle.dumps)

    def __getitem__(self, key: K) -> V:
        item = self.db.get(self.ser_key(key))
        if item is None:
            raise KeyError(key)
        return self.deser_value(item)

    def __setitem__(self, key: K, value: V):
        self.db.put(self.ser_key(key), self.ser_value(value))

    def __delitem__(self, key: K):
        self.db.delete(self.ser_key(key))

    def __iter__(self) -> Iterator[K]:
        it = self.db.iterkeys()
        it.seek_to_first()
        return (self.deser_key(key) for key in it)

    def __len__(self):
        it = self.db.iterkeys()
        it.seek_to_first()
        return sum(1 for _ in it)

    def __contains__(self, key):
        return self.db.get(self.ser_key(key)) is not None

    def keys(self) -> Iterator[K]:
        it = self.db.iterkeys()
        it.seek_to_first()
        return (self.deser_key(key) for key in it)

    def values(self) -> Iterator[V]:
        it = self.db.itervalues()
        it.seek_to_first()
        return (self.deser_value(value) for value in it)

    def get(self, key: K, default: Optional[V] = None):
        item = self.db.get(self.ser_key(key))
        if item is None:
            return default
        return self.deser_value(item)

    def pop(self, key: K, default=None):
        serkey = self.ser_key(key)
        item = self.db.get(serkey)
        if item is None:
            return default
        self.db.delete(serkey)
        return self.deser_value(item)

    def close(self):
        self.db.close()
        del self.db

    def open(self):
        self.db = DB(
            os.path.join(self.dbpath, "primary"),
            Options(**self.db_options),
            **self.kwargs
        )

    def cache_dict(self):
        return CacheDict(self)



from pathlib import Path
import shutil
from hugedict.deprecated.rocksdb import RocksDBDict
from multiprocessing import Process, Queue, Pipe
from loguru import logger


tempdir = Path("/tmp/hugedict_test")
shutil.rmtree(tempdir, ignore_errors=True)
tempdir.mkdir(parents=True, exist_ok=True)

db = RocksDBDict(dbpath="/tmp/hugedict_test/test.db", create_if_missing=True)
db["aaa"] = "10"


def process1(conn):
    assert conn.recv() == "t0"

    logger.info("[process1] online")
    logger.info("[process1] see key aaa = {}", db.get("aaa", None))

    db["bbb"] = "20"
    logger.info("[process1] see key bbb = {}", db.get("bbb", None))
    logger.info("[process1] time: t1")
    conn.send("t1")


def process2(conn):
    assert conn.recv() == "t0"

    logger.info("[process2] online")
    logger.info("[process2] see key aaa = {}", db.get("aaa", None))

    assert conn.recv() == "t1"

    logger.info("[process2] see key bbb = {}", db.get("bbb", None))


logger.info("[main] time: t0")

p1conn_pp, p1conn_cc = Pipe()
p2conn_pp, p2conn_cc = Pipe()
p1 = Process(target=process1, args=(p1conn_cc,))
p1.start()

p2 = Process(target=process2, args=(p2conn_cc,))
p2.start()

p1conn_pp.send("t0")
p2conn_pp.send("t0")

assert p1conn_pp.recv() == "t1"
p2conn_pp.send("t1")

p1.join()
p2.join()



from multiprocessing import Pool
from typing import Any, Callable, Optional, Type
import orjson
import inspect
from inspect import signature
from hugedict.v1.mrsw_rocksdb import SecondarySyncedRocksDBDict
from hugedict.v1.rocksdb import RocksDBDict
from hugedict.types import Fn
from loguru import logger


class ParallelFnWrapper:
    def __init__(self, fn: Fn, ignore_error=False):
        self.fn = fn
        # method of instances won't have `self` parameter, so we do not need to worry about it
        if hasattr(fn, "__self__") and isinstance(
            fn.__self__, (SecondaryRocksDBCacheFn, LazyRocksDBCacheFn)
        ):
            fn_params = signature(fn.__self__.fn).parameters
        else:
            fn_params = signature(fn).parameters
        self.spread_fn_args = len(fn_params) > 1
        self.ignore_error = ignore_error

    def run(self, args):
        idx, r = args
        try:
            if self.spread_fn_args:
                r = self.fn(*r)
            else:
                r = self.fn(r)
            return idx, r
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            if self.ignore_error:
                return idx, None
            raise

    def run_no_return(self, args):
        idx, r = args
        try:
            if self.spread_fn_args:
                r = self.fn(*r)
            else:
                r = self.fn(r)
            return idx, None
        except:
            logger.error(f"[ParallelMap] Error while process item {idx}")
            if self.ignore_error:
                return idx, None
            raise


class SecondaryRocksDBCacheFn:
    def __init__(
        self,
        db: SecondarySyncedRocksDBDict[bytes, bytes],
        fn: Fn,
        namespace: str = "",
    ):
        self.db = db
        self.namespace = namespace

        self.fn = fn
        self.fn_name = fn.__name__

    def run(self, *args, **kwargs):
        key = orjson.dumps((self.fn_name, args, kwargs))

        if key not in self.db:
            output = self.fn(*args, **kwargs)
            self.db[key] = output
            return output

        return self.db[key]


class LazyRocksDBCacheFn:
    def __init__(
        self,
        db_class: Type[RocksDBDict],
        db_args: dict,
        fn: Fn,
        namespace: str = "",
        key: Optional[Callable[[str, tuple, dict], bytes]] = None,
    ):
        self.db_class = db_class
        self.db_args = db_args
        self.db: Optional[RocksDBDict] = None
        self.namespace = namespace

        self.fn = fn
        self.fn_name = (
            namespace + ":" + fn.__name__ if len(namespace) > 0 else fn.__name__
        )
        self.key = key or LazyRocksDBCacheFn.default_key

    def run(self, *args, **kwargs):
        key = self.key(self.fn_name, args, kwargs)
        if self.db is None:
            self.db = self.db_class(**self.db_args)

        if key not in self.db:
            output = self.fn(*args, **kwargs)
            self.db[key] = output
            return output

        return self.db[key]

    @staticmethod
    def default_key(fn_name, args, kwargs):
        return orjson.dumps((fn_name, args, kwargs))


class CacheFnKey:
    @staticmethod
    def single_str_arg(fn_name: str, args: tuple, kwargs: dict):
        return args[0].encode()



import os
from typing import Dict, ItemsView, Iterator, KeysView, MutableMapping, ValuesView
from hugedict.types import K, V, HugeMutableMapping
from copy import copy


class CacheDict(HugeMutableMapping[K, V]):
    def __init__(self, mapping: HugeMutableMapping[K, V]) -> None:
        self.mapping = mapping
        self._cache: Dict[K, V] = {}

    def __getitem__(self, key: K) -> V:
        if key not in self._cache:
            self._cache[key] = self.mapping[key]
        return self._cache[key]

    def __setitem__(self, key: K, value: V):
        self.mapping[key] = value

    def __delitem__(self, key: K):
        if key in self._cache:
            del self._cache[key]
        del self.mapping[key]

    def __iter__(self) -> Iterator[K]:
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __contains__(self, key):
        if key in self._cache:
            return True
        return key in self.mapping

    def keys(self) -> KeysView[K]:
        return self.mapping.keys()

    def values(self) -> ValuesView[V]:
        return self.mapping.values()

    def items(self) -> ItemsView[K, V]:
        return self.mapping.items()

    def get(self, key: K, default=None):
        if key in self._cache:
            return self._cache[key]
        return self.mapping.get(key, default)

    def cache(self) -> "CacheDict":
        cache = CacheDict(self.mapping)
        # Return a clone of this object so that subsequence cache won't affect this object.
        cache._cache = copy(self._cache)
        return cache



from functools import partial
import time
from multiprocessing import Process, freeze_support
from pathlib import Path
from hugedict.hugedict.rocksdb import Options, SecondaryDB, primary_db, stop_primary_db
import sys


bench_dir = Path(__file__).parent.parent / "benchmarks"
dbpath = f"{bench_dir}/tempdir/rust-rocksdb.db"
url = "ipc:///tmp/test.ipc"
print(url)

if sys.argv[1] == "primary":
    primary_db(url, dbpath, Options())
elif sys.argv[1] == "stop":
    stop_primary_db(url)
else:
    db = SecondaryDB(
        url,
        dbpath,
        f"{dbpath}/secondaries/0",
        Options(),
        deser_value=partial(str, encoding="utf-8"),
        ser_value=str.encode,
    )
    db["P58"] = "screenwriter"
    print(db[sys.argv[1]])



from abc import abstractmethod
from loguru import logger
from multiprocessing import shared_memory, resource_tracker
from multiprocessing.managers import BaseManager, SharedMemoryManager
from multiprocessing.shared_memory import SharedMemory
from pathlib import Path
from typing import Callable, Optional, Union
from hugedict.v1.mrsw_rocksdb import (
    BloomFilterArgs,
    IProxyDict,
    PrimarySyncedRocksDBDict,
)
from hugedict.types import K, V

MB = 1024 * 1024


class MyManager(BaseManager):
    pass


MySharedMemoryManager = SharedMemoryManager


class ISharedMemoryProxyDictServer(IProxyDict):
    @abstractmethod
    def create_shared_memory(self, size: int) -> str:
        pass

    @abstractmethod
    def unlink_shared_memory(self, name: str):
        pass

    @abstractmethod
    def unlink_all_shared_memories(self):
        """Destroy all shared memory it has created."""
        pass

    @abstractmethod
    def raw_indirect_set(self, key: bytes, shared_mem_name: str, value_size: int):
        pass


class SharedMemoryDictClient(IProxyDict):
    """A client that interaction with the real proxy dict via shared memory.

    Note that we must delay creating the shared memory until the first time we actually
    need to use it to ensure that the shared memory is not appeared in the resource
    tracker process of the server. This is because of the current problematic implementation
    of shared memory in python that a shared memory is tracked twice (please see: https://bugs.python.org/issue39959).
    """

    def __init__(
        self,
        server: ISharedMemoryProxyDictServer,
        min_increase_size: int = int(1 * MB),
    ):
        self.server = server
        self.shared_mem = None
        self.min_increase_size = min_increase_size

    def __del__(self):
        if self.shared_mem is not None:
            self.shared_mem.close()
            self.server.unlink_shared_memory(self.shared_mem.name)

    def raw_set(self, key: bytes, value: bytes):
        # save data to shared memory
        value_size = len(value)
        if self.shared_mem is None or self.shared_mem.size < value_size:
            new_size = max(
                round(value_size / self.min_increase_size) * self.min_increase_size,
                self.min_increase_size,
            )
            if self.shared_mem is not None:
                self.shared_mem.close()
                self.server.unlink_shared_memory(self.shared_mem.name)
            self.shared_mem = SharedMemory(
                name=self.server.create_shared_memory(new_size), create=False
            )
            # resource tracker register shared memory twice - unregister to avoid warnings
            # see more: (the first one is related to our problem but declared as duplicated as the one below)
            # + https://bugs.python.org/issue39959
            # + https://bugs.python.org/issue38119
            resource_tracker.unregister(self.shared_mem._name, "shared_memory")  # type: ignore

        self.shared_mem.buf[:value_size] = value
        self.server.raw_indirect_set(key, str(self.shared_mem.name), value_size)

    def raw_has(self, key: bytes) -> bool:
        return self.server.raw_has(key)

    def raw_get(self, key: bytes) -> Optional[bytes]:
        return self.server.raw_get(key)


class SharedMemoryPrimarySyncedRocksDBDict(
    PrimarySyncedRocksDBDict[K, V], ISharedMemoryProxyDictServer
):
    def __init__(
        self,
        dbpath: Union[Path, str],
        create_if_missing=False,
        deser_key: Callable[[bytes], K] = None,
        deser_value: Callable[[bytes], V] = None,
        ser_key: Callable[[K], bytes] = None,
        ser_value: Callable[[V], bytes] = None,
        bloomfilter_args: BloomFilterArgs = None,
        enable_bloomfilter: bool = True,
    ):
        super().__init__(
            dbpath,
            create_if_missing,
            deser_key,
            deser_value,
            ser_key,
            ser_value,
            bloomfilter_args,
            enable_bloomfilter,
        )

        self.shared_mems = {}

    def create_shared_memory(self, size: int) -> str:
        sm = SharedMemory(create=True, size=size)
        self.shared_mems[sm.name] = sm
        return sm.name

    def unlink_shared_memory(self, name: str):
        self.shared_mems[name].unlink()
        del self.shared_mems[name]

    def unlink_all_shared_memories(self):
        names = list(self.shared_mems.keys())
        for name in names:
            self.unlink_shared_memory(name)

    def raw_indirect_set(self, key: bytes, shared_mem_name: str, value_size: int):
        value = self.shared_mems[shared_mem_name].buf[:value_size]
        self.raw_set(key, bytes(value))


MyManager.register("PrimarySyncedRocksDBDict", PrimarySyncedRocksDBDict)
MyManager.register(
    "SharedMemoryPrimarySyncedRocksDBDict", SharedMemoryPrimarySyncedRocksDBDict
)



import os
import shutil
from typing import Generator
from glob import glob
from pathlib import Path

import pytest

from hugedict.hugedict.rocksdb import load, Options


@pytest.fixture()
def wdprops(resource_dir: Path, tmp_path: Path) -> Generator[Path, None, None]:
    dbpath = tmp_path / f"wdprops.db"

    load(
        str(dbpath),
        Options(create_if_missing=True),
        glob(f"{resource_dir}/wdprops/*.tsv"),
        {
            "record_type": {"type": "tabsep", "key": None, "value": None},
            "is_sorted": False,
        },
        verbose=False,
        compact=True,
    )

    yield dbpath

    shutil.rmtree(dbpath)



"""Loading big files into RocksDB
"""
import time
import bz2
from dataclasses import dataclass
import gzip
from math import ceil
import os
from pathlib import Path
import orjson
from rocksdb import DB, WriteBatch  # type: ignore
from enum import Enum
from typing import Any, Callable, List, Optional, Tuple, Union, cast
from multiprocessing.managers import SharedMemoryManager, SyncManager
from multiprocessing.shared_memory import SharedMemory, ShareableList
from multiprocessing import Lock, Pool, Queue, Process
from tqdm import tqdm


class FileFormat(str, Enum):
    # each line is a json list of two items, key is extracted from the first item while value is extracted from the second item
    tuple2 = "tuple2"
    # each line is a json object, key and value are extracted from the object
    jsonline = "jsonline"
    # tab separated format, raw key and value must not contain tab and newline characters such as \r\n.
    tabsep = "tabsep"


@dataclass
class FileReaderArgs:
    infile: Path
    format: FileFormat
    key_fn: Callable[[Any], bytes]
    value_fn: Callable[[Any], bytes]
    shm_pool: List[SharedMemory]
    shm_reserved: ShareableList


def init_pool(l):
    global lock
    lock = l


def load_single_file(
    db: DB,
    infile: Union[str, Path],
    format: FileFormat,
    key_fn: Callable[[Any], bytes],
    value_fn: Callable[[Any], bytes],
    verbose: bool = True,
):
    args = FileReaderArgs(
        infile=Path(infile),
        format=format,
        key_fn=key_fn,
        value_fn=value_fn,
        shm_pool=[],
        shm_reserved=None,  # type: ignore
    )
    lst = read_file(args)
    batch_size = 100000

    wb = WriteBatch()
    for k, v in tqdm(lst, desc="writing", disable=not verbose):
        wb.put(k, v)
        if wb.count() >= batch_size:
            db.write(wb, disable_wal=True)
            wb = WriteBatch()
    if wb.count() > 0:
        db.write(wb, disable_wal=True)
    return db


def load(
    db: DB,
    infiles: Union[List[str], List[Path], List[Union[str, Path]]],
    format: FileFormat,
    key_fn: Callable[[Any], bytes],
    value_fn: Callable[[Any], bytes],
    n_processes: Optional[int] = None,
    shm_mem_limit_mb: int = 128,
    shm_mem_ratio: int = 4,
    verbose: bool = True,
):
    """Load files into rocksdb database"""
    infiles = sorted(infiles)
    if n_processes is None:
        n_cpus = os.cpu_count()
        if n_cpus is None:
            raise Exception("Cannot determine number of CPUs")
        n_processes = min(min(4, n_cpus - 4), n_cpus)

    with cast(Any, SharedMemoryManager()) as smm, tqdm(
        desc="loading", total=len(infiles), disable=not verbose
    ) as pbar:
        n_shared_mems = min(ceil(n_processes * shm_mem_ratio), len(infiles))
        shm_pool = [
            smm.SharedMemory(1024 * 1024 * shm_mem_limit_mb)
            for _ in range(n_shared_mems)
        ]
        shm_reserved = smm.ShareableList([False for _ in range(n_shared_mems)])
        lock = Lock()
        with Pool(
            processes=n_processes, initializer=init_pool, initargs=(lock,)
        ) as pool:
            inputs = [
                FileReaderArgs(
                    infile=Path(infile),
                    format=format,
                    key_fn=key_fn,
                    value_fn=value_fn,
                    shm_pool=shm_pool,
                    shm_reserved=shm_reserved,
                )
                for infile in infiles
            ]
            n_records = 0
            for shm_index in pool.imap_unordered(process_fn, inputs):
                if not shm_reserved[shm_index]:
                    raise Exception(f"Shared memory {shm_index} not reserved")

                pbar.update(1)
                # read and release shared memory
                shm = shm_pool[shm_index]
                # output = read_from_shm(shm)
                pbar.set_postfix(records=n_records, msg="reading batch to db")
                wb = read_from_shm_writebatch(shm)
                lock.acquire()
                try:
                    shm_reserved[shm_index] = False
                finally:
                    lock.release()

                n_records += wb.count()
                # wb = WriteBatch()
                # for k, v in output:
                #     wb.put(k, v)
                #     n_records += 1
                #     pbar.set_postfix(records=n_records, msg="creating batch")

                pbar.set_postfix(records=n_records, msg="writing batch to db")
                db.write(wb, disable_wal=True)
                pbar.set_postfix(records=n_records, msg="wrote batch to db")

    return db


def process_fn(args: FileReaderArgs) -> int:
    """Read file and write list of key-value pairs to shared memory"""
    global lock
    outputs = read_file(args)

    shm_idx = None
    # wait for max 10 minutes for shared memory to be available
    for j in range(600):
        for i in range(len(args.shm_pool)):
            if args.shm_reserved[i]:
                continue

            lock.acquire()
            try:
                if not args.shm_reserved[i]:
                    args.shm_reserved[i] = True
                else:
                    continue
            finally:
                lock.release()
            shm_idx = i
            break
        if shm_idx is None:
            time.sleep(1)
        else:
            break

    if shm_idx is None:
        raise Exception("No shared memory available")

    # write_to_shm(args.shm_pool[shm_idx], outputs)
    write_to_shm_writebatch(args.shm_pool[shm_idx], outputs)
    return shm_idx


def read_file(args: FileReaderArgs) -> List[Tuple[bytes, bytes]]:
    if args.infile.name.endswith(".gz"):
        open_fn = gzip.open
    elif args.infile.name.endswith(".bz2"):
        open_fn = bz2.open
    else:
        open_fn = open

    with open_fn(str(args.infile), "rb") as f:
        key = args.key_fn
        value = args.value_fn
        outputs = []
        if args.format == FileFormat.jsonline:
            for line in f:
                r = orjson.loads(line)
                outputs.append((key(r), value(r)))
        elif args.format == FileFormat.tuple2:
            for line in f:
                k, v = orjson.loads(line)
                outputs.append((key(k), value(v)))
        elif args.format == FileFormat.tabsep:
            for line in f:
                k, v = line.rstrip(b"\r\n").split(b"\t", 1)
                outputs.append((key(k), value(v)))
        else:
            raise Exception(f"Unknown format: {format}")
    return outputs


def write_to_shm_writebatch(shm: SharedMemory, lst: List[Tuple[bytes, bytes]]):
    wb = WriteBatch()
    for k, v in lst:
        wb.put(k, v)
    content = wb.data()
    size = len(content)
    if shm.size < size:
        raise Exception(f"Shared memory too small: {shm.size} < {size}")
    shm.buf[:4] = size.to_bytes(4, "little")
    shm.buf[4 : 4 + size] = content


def read_from_shm_writebatch(shm: SharedMemory) -> WriteBatch:
    size = int.from_bytes(shm.buf[:4], "little")
    return WriteBatch(bytes(shm.buf[4 : 4 + size]))


def write_to_shm(shm: SharedMemory, lst: List[Tuple[bytes, bytes]]):
    n_items = len(lst) * 2
    size = 4 * n_items + sum(len(k) + len(v) for k, v in lst) + 4
    if shm.size < size:
        raise Exception(f"Shared memory too small: {shm.size} < {size}")
    shm.buf[:4] = n_items.to_bytes(4, "little")
    i = 4
    for slst in lst:
        for item in slst:
            shm.buf[i : i + 4] = len(item).to_bytes(4, "little")
            shm.buf[i + 4 : i + 4 + len(item)] = item
            i += len(item) + 4


def read_from_shm(shm: SharedMemory) -> List[Tuple[bytes, bytes]]:
    lst = []
    n_items = int.from_bytes(shm.buf[:4], "little")
    i = 4
    for _ in range(n_items):
        size = int.from_bytes(shm.buf[i : i + 4], "little")
        k = bytes(shm.buf[i + 4 : i + 4 + size])
        i += size + 4

        size = int.from_bytes(shm.buf[i : i + 4], "little")
        v = bytes(shm.buf[i + 4 : i + 4 + size])
        i += size + 4

        lst.append((k, v))
    return lst



import shutil
from hugedict.parallel import Parallel
from hugedict.v1.parallel.manager import MB
from hugedict.v1.parallel.parallel import Compressing
from loguru import logger
import orjson
import sm.misc as M

# Download the data in here: https://drive.google.com/file/d/1GPi19KEpBgf-AoQ2XFHXYEOUdFcg47tL/view?usp=sharing
# and put it to /tmp/test-break-syncmanager.jl.gz
logger.info("Loading data")
records = M.deserialize_jl("/tmp/test-break-syncmanager.jl.gz", n_lines=100)
records = {orjson.dumps(r["key"]).decode(): r["value"] for r in records}
logger.info("Start testing")

pp = Parallel(enable_shared_memory=True, min_increase_shm_mb_size=5)

shutil.rmtree("/tmp/hugedict/test2.db", ignore_errors=True)


@pp.cache_func("/tmp/hugedict/test2.db", compress=Compressing.NoCompression)
def run(q):
    global records
    return records[q]


def run2(q):
    run(q)
    return 5


output = pp.foreach(run2, sorted(records.keys()), show_progress=True, is_parallel=True)
print("FINISH!!")



import functools
import gzip
import threading
import zstandard as zstd
import pickle
from typing import Any, Callable
from hugedict.types import T


container = threading.local()
container.zstd_3_compressor = zstd.ZstdCompressor(level=3)
container.zstd_6_compressor = zstd.ZstdCompressor(level=6)
container.zstd_decompressor = zstd.ZstdDecompressor()


def init_container(container):
    if not hasattr(container, "zstd_decompressor"):
        container.zstd_3_compressor = zstd.ZstdCompressor(level=3)
        container.zstd_6_compressor = zstd.ZstdCompressor(level=6)
        container.zstd_decompressor = zstd.ZstdDecompressor()


def identity(x):
    return x


def compress_pyobject(x):
    return gzip.compress(pickle.dumps(x), mtime=0)


def decompress_pyobject(x):
    return pickle.loads(gzip.decompress(x))


def zstd3_compress(x):
    global container
    init_container(container)
    return container.zstd_3_compressor.compress(x)


def zstd6_compress(x):
    global container
    init_container(container)
    return container.zstd_6_compressor.compress(x)


def zstd_decompress(x):
    global container
    init_container(container)
    return container.zstd_decompressor.decompress(x)


def compress_zstd3_pyobject(x):
    global container
    init_container(container)
    return container.zstd_3_compressor.compress(pickle.dumps(x))


def compress_zstd6_pyobject(x):
    global container
    init_container(container)
    return container.zstd_6_compressor.compress(pickle.dumps(x))


def decompress_zstd_pyobject(x):
    global container
    init_container(container)
    return pickle.loads(container.zstd_decompressor.decompress(x))


def compress_custom(ser_fn: Callable[[T], bytes]) -> Callable[[T], bytes]:
    @functools.wraps(ser_fn)
    def wrapper(x):
        return gzip.compress(ser_fn(x), mtime=0)

    return wrapper


def decompress_custom(deser_fn: Callable[[bytes], T]) -> Callable[[bytes], T]:
    @functools.wraps(deser_fn)
    def wrapper(x):
        return deser_fn(gzip.decompress(x))

    return wrapper


def zstd3_compress_custom(ser_fn: Callable[[T], bytes]) -> Callable[[T], bytes]:
    @functools.wraps(ser_fn)
    def wrapper(x):
        global container
        init_container(container)
        return container.zstd_3_compressor.compress(ser_fn(x))

    return wrapper


def zstd6_compress_custom(ser_fn: Callable[[T], bytes]) -> Callable[[T], bytes]:
    @functools.wraps(ser_fn)
    def wrapper(x):
        global container
        init_container(container)
        return container.zstd_6_compressor.compress(ser_fn(x))

    return wrapper


def zstd_decompress_custom(deser_fn: Callable[[bytes], T]) -> Callable[[bytes], T]:
    @functools.wraps(deser_fn)
    def wrapper(x):
        global container
        init_container(container)
        return deser_fn(container.zstd_decompressor.decompress(x))

    return wrapper


class Chain2:
    def __init__(self, g, f):
        self.g = g
        self.f = f

    def exec(self, args):
        return self.g(self.f(args))

    def __call__(self, args):
        return self.g(self.f(args))


class Chain3:
    def __init__(self, h, g, f):
        self.h = h
        self.g = g
        self.f = f

    def exec(self, args):
        return self.h(self.g(self.f(args)))



from pathlib import Path
import time, tempfile, pytest

from hugedict.v1.parallel import Parallel


def heavy_computing(seconds: float):
    time.sleep(seconds)
    return seconds * 2


@pytest.mark.parametrize(
    "enable_bloomfilter,enable_shared_memory",
    [
        (enable_bloomfilter, enable_shared_memory)
        for enable_bloomfilter in [True, False]
        for enable_shared_memory in [True, False]
    ],
)
@pytest.mark.skip()
def test_parallel(enable_bloomfilter: bool, enable_shared_memory: bool):
    pp = Parallel(
        enable_bloomfilter=enable_bloomfilter, enable_shared_memory=enable_shared_memory
    )

    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = Path(tempdir)
        fn = pp.cache_func(tempdir / "test.db")(heavy_computing)
        output = pp.map(fn, [0.5, 1, 0.7, 0.3, 0.6], n_processes=3)

        assert output == [1, 2, 1.4, 0.6, 1.2]



import time
from pathlib import Path

from hugedict.parallel import Parallel


def heavy_computing(seconds: float):
    time.sleep(seconds)
    return seconds * 2


def test_parallel(tmp_path: Path, url: str):
    pp = Parallel()

    fn = pp.cache_func(tmp_path / "test.db", url=url)(heavy_computing)
    output = pp.map(fn, [0.5, 1, 0.7, 0.3, 0.6], n_processes=3)

    assert output == [1, 2, 1.4, 0.6, 1.2]



import enum, gc
import functools
from logging import disable
import os
from contextlib import contextmanager
from pathlib import Path
import shutil
from uuid import uuid4
from functools import partial
from multiprocessing.pool import Pool, ThreadPool
from multiprocessing.managers import BaseManager, SharedMemoryManager, SyncManager
from operator import itemgetter
import pickle
from typing import Any, Callable, List, Optional, Union
from hugedict.v1.mrsw_rocksdb import SecondarySyncedRocksDBDict
from hugedict.v1.parallel.fn_wrapper import (
    LazyRocksDBCacheFn,
    ParallelFnWrapper,
    CacheFnKey,
)
from hugedict.misc import (
    compress_zstd6_pyobject,
    decompress_zstd_pyobject,
    identity,
    compress_pyobject,
    decompress_pyobject,
)
from hugedict.v1.parallel.manager import (
    MB,
    MyManager,
    SharedMemoryDictClient,
)
from hugedict.v1.rocksdb import RocksDBDict
from hugedict.types import F, Fn
from loguru import logger
from tqdm import tqdm


class Compressing(enum.Flag):
    NoCompression = enum.auto()
    CompressKey = enum.auto()
    CompressValue = enum.auto()
    CompressAll = CompressValue | CompressKey


class Parallel:
    def __init__(
        self,
        enable_bloomfilter=True,
        enable_shared_memory=False,
        min_increase_shm_mb_size=2,
    ):
        self._cache: List[LazyRocksDBCacheFn] = []
        self._cache_primary: list = []
        self.enable_shared_memory = enable_shared_memory
        self.enable_bloomfilter = enable_bloomfilter
        self.min_increase_shm_size = int(min_increase_shm_mb_size * MB)

    @contextmanager
    def init_cache(self):
        try:
            for cache in self._cache:
                if cache.db is not None:
                    cache.db.close()
                    cache.db = None
            yield
        finally:
            for cache in self._cache:
                if cache.db is not None:
                    cache.db.close()
                    cache.db = None

    @contextmanager
    def switch_mrsw_cache(self):
        with MyManager() as manager:
            lst_db_args = []
            servers = []
            try:
                for cache in self._cache:
                    assert cache.db is None, "Freshly start db"
                    # clean previous files except primary db
                    dbpath = cache.db_args["dbpath"]
                    if dbpath.exists():
                        for file in dbpath.iterdir():
                            if file.name != "primary":
                                if file.is_dir():
                                    shutil.rmtree(file)
                                else:
                                    file.unlink()

                    if self.enable_shared_memory:
                        server = manager.SharedMemoryPrimarySyncedRocksDBDict(  # type: ignore
                            enable_bloomfilter=self.enable_bloomfilter, **cache.db_args
                        )
                        servers.append(server)
                        primary = SharedMemoryDictClient(
                            server, min_increase_size=self.min_increase_shm_size
                        )
                    else:
                        primary = manager.PrimarySyncedRocksDBDict(  # type: ignore
                            enable_bloomfilter=self.enable_bloomfilter, **cache.db_args
                        )
                    cache.db_class = SecondarySyncedRocksDBDict
                    lst_db_args.append(cache.db_args.copy())
                    del cache.db_args["create_if_missing"]
                    cache.db_args["primary"] = primary
                    cache.db_args["enable_bloomfilter"] = self.enable_bloomfilter
                    cache.db_args["secondary_name"] = str(uuid4()).replace("-", "")
                yield
            finally:
                for server in servers:
                    server.unlink_all_shared_memories()
                for cache, db_args in zip(self._cache, lst_db_args):
                    cache.db_class = RocksDBDict
                    cache.db_args = db_args

    def map(
        self,
        fn: Fn,
        inputs: list,
        show_progress=False,
        progress_desc="",
        is_parallel=True,
        use_threadpool=False,
        n_processes: Optional[int] = None,
        ignore_error: bool = False,
    ) -> List[Any]:
        """Execute a map function over each input in parallel"""
        with self.init_cache():
            if not is_parallel:
                wrapped_fn = ParallelFnWrapper(fn, ignore_error=False).run
                return [
                    wrapped_fn((i, item))[1]
                    for i, item in tqdm(
                        enumerate(inputs),
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    )
                ]

            if use_threadpool:
                # it won't break rocksdb when using threadpool because of GIL.
                with ThreadPool(processes=n_processes) as pool:
                    iter = pool.imap_unordered(
                        ParallelFnWrapper(fn, ignore_error).run,
                        enumerate(inputs),
                    )
                    results = []
                    for result in tqdm(
                        iter,
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    ):
                        results.append(result)
                    results.sort(key=itemgetter(0))
            else:
                # have to switch to multi read single write mode
                with self.switch_mrsw_cache():
                    # start a pool of processes
                    with Pool(processes=n_processes) as pool:
                        iter = pool.imap_unordered(
                            ParallelFnWrapper(fn, ignore_error).run,
                            enumerate(inputs),
                        )
                        results = []
                        for result in tqdm(
                            iter,
                            total=len(inputs),
                            desc=progress_desc,
                            disable=not show_progress,
                        ):
                            results.append(result)
                        results.sort(key=itemgetter(0))

        return [v for i, v in results]

    def foreach(
        self,
        fn: Fn,
        inputs: list,
        show_progress=False,
        progress_desc="",
        is_parallel=True,
        use_threadpool=False,
        n_processes: Optional[int] = None,
        ignore_error: bool = False,
    ) -> None:
        """Execute a map function over each input in parallel"""
        with self.init_cache():
            if not is_parallel:
                for item in tqdm(
                    inputs,
                    total=len(inputs),
                    desc=progress_desc,
                    disable=not show_progress,
                ):
                    fn(item)
                return

            if use_threadpool:
                # it won't break rocksdb when using threadpool because of GIL.
                with ThreadPool(processes=n_processes) as pool:
                    iter = pool.imap_unordered(
                        ParallelFnWrapper(fn, ignore_error).run_no_return,
                        enumerate(inputs),
                    )
                    results = []
                    for result in tqdm(
                        iter,
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    ):
                        results.append(result)
                    return

            # have to switch to multi read single write mode
            with self.switch_mrsw_cache():
                # start a pool of processes
                with Pool(processes=n_processes) as pool:
                    iter = pool.imap_unordered(
                        ParallelFnWrapper(fn, ignore_error).run_no_return,
                        enumerate(inputs),
                    )
                    results = []
                    for result in tqdm(
                        iter,
                        total=len(inputs),
                        desc=progress_desc,
                        disable=not show_progress,
                    ):
                        results.append(result)
                    return

    def cache_func(
        self,
        dbpath: Union[Path, str],
        namespace: str = "",
        compress=Compressing.NoCompression,
        key: Optional[Callable[[str, tuple, dict], bytes]] = None,
    ) -> Callable[[F], F]:
        """Cache a function (only work when using with Parallel object)

        Args:
            dbpath: path to the db
            namespace: namespace of the function, used to distinguish if you want to cache multiple functions in the same db
            compress: whether to explicitly compress key or value of a record
            key: a function to generate key from (namespace + ":" + func_name, func_args, func_kwargs)
        """
        for cache in self._cache:
            assert cache.db_args["dbpath"] != dbpath, "dbpath must be unique"

        def wrapper_fn(func: F) -> F:
            if compress & Compressing.CompressKey:
                ser_key, deser_key = compress_pyobject, decompress_pyobject
            else:
                ser_key, deser_key = identity, identity

            if compress & Compressing.CompressValue:
                ser_value, deser_value = (
                    compress_zstd6_pyobject,
                    decompress_zstd_pyobject,
                )
            else:
                ser_value, deser_value = pickle.dumps, pickle.loads

            cache = LazyRocksDBCacheFn(
                db_class=RocksDBDict,
                db_args=dict(
                    dbpath=Path(dbpath),
                    create_if_missing=True,
                    deser_key=deser_key,
                    ser_key=ser_key,
                    deser_value=deser_value,
                    ser_value=ser_value,
                ),
                fn=func,
                namespace=namespace,
                key=key,
            )

            self._cache.append(cache)
            return cache.run  # type: ignore

        return wrapper_fn



# Prerequisites:
# - install sem-desc: `pip install sem-desc`
# - data to test with in the ./data directory (e.g., ./data/wdentities/part-*.tsv.gz)

from dataclasses import dataclass
import multiprocessing
import shutil
from pathlib import Path
from typing import Optional

from hugedict.v1.loader import FileFormat, load
from rocksdb import DB, Options  # type: ignore
from sm.prelude import M


multiprocessing.set_start_method("fork")

bench_dir = Path(__file__).parent
shutil.rmtree(bench_dir / "tempdir", ignore_errors=True)
(bench_dir / "tempdir").mkdir(exist_ok=True)
(bench_dir / "tempdir" / "python-rocksdb.db").mkdir(exist_ok=True)

infiles = [str(x) for x in (bench_dir / "data/wdentities").glob("part-*.tsv.gz")]

with M.Timer().watch_and_report("python-rocksdb"):
    db = DB(
        str(bench_dir / "tempdir" / "python-rocksdb.db" / "primary"),
        Options(create_if_missing=True),
    )
    load(
        db,
        infiles,
        FileFormat.tabsep,
        key_fn=M.identity_func,
        value_fn=M.identity_func,
        verbose=False,
    )
    db.compact_range()


from hugedict.hugedict import rocksdb


with M.Timer().watch_and_report("rust-rocksdb"):
    rocksdb.load(
        str(bench_dir / "tempdir" / "rust-rocksdb.db"),
        rocksdb.Options(create_if_missing=True),
        infiles,
        format={
            "record_type": {"type": "tabsep", "key": None, "value": None},
            "is_sorted": False,
        },
        verbose=True,
        compact=True,
    )



# Prerequisites:
# - install sem-desc (for `sm`): `pip install sem-desc`
# - install kgdata: `pip install kgdata`
# - and other packages

import random
from pathlib import Path

import orjson
from kgdata.wikidata.db import get_entity_db
from sm.prelude import M
from tqdm import tqdm
from loguru import logger

data_dir = Path(__file__).parent / "data"

# generate benchmark data
random.seed(10)

db = get_entity_db("/workspace/sm-dev/data/wikidata/20211213/databases/wdentities.db")
n = 3000
part_size = 100
kvs = []

counter = 1
for i in tqdm(range(n), desc="load records"):
    for i in range(100):
        ent = db.get(f"Q{counter}", None)
        if ent is not None:
            break
        counter = counter + 1
    else:
        # we try it 100 times, if we don't find the entity, we hit the end of the db
        break

    kvs.append(ent.id.encode() + b"\t" + orjson.dumps(ent.to_dict()))
    counter += random.randint(1, 20)

random.shuffle(kvs)

logger.info("Creating wikidata dataset of {} records", len(kvs))

(data_dir / "wdentities").mkdir(exist_ok=True, parents=True)
counter = 0
for i in tqdm(range(0, n, part_size), desc="write records to disk"):
    file = data_dir / "wdentities" / f"part-{counter:04d}.tsv.gz"
    if len(kvs[i : i + part_size]) == 0:
        break
    M.serialize_byte_lines(kvs[i : i + part_size], file)
    counter += 1



from __future__ import print_function
import os, subprocess, re
from typing import Tuple

from pyparsing import Literal


def is_python_home(dir: str):
    return any(
        os.path.exists(os.path.join(dir, "bin", name))
        for name in ["python", "python2", "python3"]
    )


def get_python(home: str):
    for name in ["python", "python2", "python3"]:
        path = os.path.join(home, "bin", name)
        if os.path.exists(path):
            return path
    raise ValueError("python not found in {}".format(home))


def get_python_version(home: str) -> Tuple[str, str]:
    """Get python's version and whether it is cpython or pypy."""
    output = subprocess.check_output([get_python(home), "-V"]).decode().strip()
    m = re.match("Python ([\d\.)]+)", output)
    assert m is not None
    version = m.group(1)
    type = "pypy" if output.find("PyPy") != -1 else "cpython"
    return version, type


homes = {}
if "PYTHON_HOME" in os.environ:
    home = os.environ["PYTHON_HOME"]
    homes[home] = get_python_version(home)[0]
elif "PYTHON_HOMES" in os.environ:
    lst = os.environ["PYTHON_HOMES"].split(":")
    for path in lst:
        if is_python_home(path):
            # is the python directory
            homes[path] = get_python_version(path)[0]
        else:
            subhomes = {}
            for subpath in os.listdir(path):
                home = os.path.join(path, subpath)
                if not is_python_home(home):
                    continue

                version, pytype = get_python_version(home)

                # do not keep different patches (only keep major.minor)
                mm_version = ".".join(version.split(".")[:2])
                subhomes[mm_version, pytype] = (home, version)

            for home, version in subhomes.values():
                homes[home] = version

if "PYTHON_VERSIONS" in os.environ:
    versions = os.environ["PYTHON_VERSIONS"].split(",")
    filtered_homes = []
    for home, home_version in homes.items():
        for version in versions:
            if home_version.startswith(version):
                filtered_homes.append(home)
                break

    homes = {h: homes[h] for h in filtered_homes}


if "MINIMUM_PYTHON_VERSION" in os.environ:
    minimum_version = [int(d) for d in os.environ["MINIMUM_PYTHON_VERSION"].split(".")]
    filtered_homes = []
    for home, home_version in homes.items():
        pyversion = home_version.split(".")
        if all(
            int(pyversion[i]) >= minimum_version[i] for i in range(len(minimum_version))
        ):
            filtered_homes.append(home)

    homes = {h: homes[h] for h in filtered_homes}

print(":".join(homes))
exit(0)



from dataclasses import dataclass
from typing import (
    Callable,
    Generic,
    Iterator,
    Literal,
    Optional,
    List,
    Tuple,
    TypedDict,
)

from hugedict.types import KP, V, HugeMutableMapping

DBCompactionStyle = Literal["level", "universal", "fifo"]
DBCompressionStyle = Literal[
    "none",
    "snappy",
    "zlib",
    "bz2",
    "lz4",
    "lz4hc",
    "zstd",
]

class PrefixExtractor: ...

def fixed_prefix(*, type: Literal["fixed_prefix"], size: int) -> PrefixExtractor: ...
def fixed_prefix_alike(
    *, type: Literal["fixed_prefix_alike"], prefix: str
) -> PrefixExtractor: ...
@dataclass
class Options:
    create_if_missing: Optional[bool] = None
    max_open_files: Optional[int] = None
    use_fsync: Optional[bool] = None
    bytes_per_sync: Optional[int] = None
    optimize_for_point_lookup: Optional[int] = None
    table_cache_numshardbits: Optional[int] = None
    max_write_buffer_number: Optional[int] = None
    write_buffer_size: Optional[int] = None
    target_file_size_base: Optional[int] = None
    min_write_buffer_number_to_merge: Optional[int] = None
    level_zero_stop_writes_trigger: Optional[int] = None
    level_zero_slowdown_writes_trigger: Optional[int] = None
    compaction_style: Optional[DBCompactionStyle] = None
    disable_auto_compactions: Optional[bool] = None
    max_background_jobs: Optional[int] = None
    max_subcompactions: Optional[int] = None
    compression_type: Optional[DBCompressionStyle] = None
    bottommost_compression_type: Optional[DBCompressionStyle] = None
    prefix_extractor: Optional[PrefixExtractor] = None

class FileFormat(TypedDict):
    record_type: RecordType
    # whether the file is sorted or not.
    is_sorted: bool

class RecordType(TypedDict):
    """
    tabsep:
        tab separated format of serialized byte key and value
        serialized key must not contain tab character
        serialized value must not contain newline character such as \r\n.
        no key or value
    ndjson:
        each line is a json object, key is required but value is optional
    tuple2:
        each line is a json list of two items key and value, key and value are optional
    """

    type: Literal["tabsep", "ndjson", "tuple2"]
    # object's attribute contains the key, None if key is the object itself.
    key: Optional[str]
    # object's attribute contains the value, None if value is the object itself.
    value: Optional[str]

def load(
    dbpath: str,
    dbopts: Options,
    infiles: List[str],
    format: FileFormat,
    verbose: bool,
    compact: bool,
) -> None:
    """Load files into rocksdb database by building SST files and ingesting them.

    Arguments:
        dbpath: path to rocksdb database
        dbopts: rocksdb options
        infiles: list of input files
        format: file format
        verbose: whether to print progress
        compact: whether to compact the database after loading
    """

class RocksDBDict(HugeMutableMapping[KP, V]):
    """A mutable mapping backed by rocksdb.

    Arguments:
        path: path to rocksdb database
        opts: rocksdb options
        deser_key: deserialize key from a memoryview
        deser_value: deserialize value from a memoryview
        ser_value: serialize value to bytes
        readonly: whether to open the database in readonly mode (can open many times)
        secondary_mode: whether to open the database in secondary mode
        secondary_path: path to secondary rocksdb database
    """

    def __init__(
        self,
        path: str,
        options: Options,
        deser_key: Callable[[memoryview], KP],
        deser_value: Callable[[memoryview], V],
        ser_value: Callable[[V], bytes],
        readonly: bool = False,
        secondary_mode: bool = False,
        secondary_path: Optional[str] = None,
    ): ...
    @property
    def deser_value(self) -> Callable[[memoryview], V]:
        """Deserialize value from a memoryview."""
    @property
    def ser_value(self) -> Callable[[V], bytes]:
        """Serialize value to bytes."""
    def _put(self, k: bytes, v: bytes):
        """Put the raw (bytes) key and value into the database."""
    def get_int_property(
        self,
        name: Literal[
            "rocksdb.num-immutable-mem-table",
            "rocksdb.mem-table-flush-pending",
            "rocksdb.compaction-pending",
            "rocksdb.background-errors",
            "rocksdb.cur-size-active-mem-table",
            "rocksdb.cur-size-all-mem-tables",
            "rocksdb.size-all-mem-tables",
            "rocksdb.num-entries-active-mem-table",
            "rocksdb.num-entries-imm-mem-tables",
            "rocksdb.num-deletes-active-mem-table",
            "rocksdb.num-deletes-imm-mem-tables",
            "rocksdb.estimate-num-keys",
            "rocksdb.estimate-table-readers-mem",
            "rocksdb.is-file-deletions-enabled",
            "rocksdb.num-snapshots",
            "rocksdb.oldest-snapshot-time",
            "rocksdb.num-live-versions",
            "rocksdb.current-super-version-number",
            "rocksdb.estimate-live-data-size",
            "rocksdb.min-log-number-to-keep",
            "rocksdb.min-obsolete-sst-number-to-keep",
            "rocksdb.total-sst-files-size",
            "rocksdb.live-sst-files-size",
            "rocksdb.base-level",
            "rocksdb.estimate-pending-compaction-bytes",
            "rocksdb.num-running-compactions",
            "rocksdb.num-running-flushes",
            "rocksdb.actual-delayed-write-rate",
            "rocksdb.is-write-stopped",
            "rocksdb.estimate-oldest-key-time",
            "rocksdb.block-cache-capacity",
            "rocksdb.block-cache-usage",
            "rocksdb.block-cache-pinned-usage",
        ],
    ) -> Optional[int]:
        """Retrieves a RocksDB property's value and cast it to an integer.

        Full list of properties that return int values could be find [here](https://github.com/facebook/rocksdb/blob/08809f5e6cd9cc4bc3958dd4d59457ae78c76660/include/rocksdb/db.h#L654-L689).
        """
    def seek_keys(self, prefix: KP) -> Iterator[KP]:
        """Seek to the first key that matches the *entire* prefix. From
        there, the itereator will continue to read pairs as long as the
        prefix extracted from `key` matches the prefix extracted from `prefix`.

        Note: for this function to always iterate over keys that match the *entire*
        prefix, set options.prefix_extractor to the length of the prefix.
        """
    def seek_items(self, prefix: KP) -> Iterator[Tuple[KP, V]]:
        """Seek to the first key that matches the *entire* prefix. From
        there, the itereator will continue to read pairs as long as the
        prefix extracted from `key` matches the prefix extracted from `prefix`.

        Note: for this function to always iterate over keys that match the *entire*
        prefix, set options.prefix_extractor to the length of the prefix.
        """
    def compact(self, start: Optional[KP], end: Optional[KP]):
        """Compact the database on the range of keys"""
    def try_catch_up_with_primary(self):
        """For secondary mode. Tries to catch up with the primary by reading as much as possible from the log files."""

def primary_db(url: str, path: str, opts: Options):
    """Start a primary rocksdb. Should be in a separated process."""

def stop_primary_db(url: str):
    """Stop the primary rocksdb."""

class SecondaryDB(Generic[KP, V]):
    """A secondary instance of rocksdb that falled back to primary if not found.

    It communicates with primary instance using nng.

    Arguments:
        url: url of the socket
        primary_path: path to primary rocksdb
        secondary_path: path to secondary rocksdb
        opts: rocksdb options
        deser_value: deserialize value from a memoryview
        ser_value: serialize value to bytes
    """

    def __init__(
        self,
        url: str,
        primary_path: str,
        secondary_path: str,
        opts: Options,
        deser_value: Callable[[memoryview], V],
        ser_value: Callable[[V], bytes],
    ): ...
    def __getitem__(self, key: KP) -> V:
        """Get key from database"""
        ...
    def __setitem__(self, key: KP, value: V) -> None:
        """Put key and value to database"""
        ...
    def __contains__(self, key: KP) -> bool:
        """Check if key is in database"""
        ...
